# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Constants for the solana ledger."""
_VERSION = "1.24.17"
_SOLANA = "solana"
TESTNET_NAME = "n/a"
DEFAULT_ADDRESS = "https://api.devnet.solana.com"
DEFAULT_CHAIN_ID = 101
DEFAULT_CURRENCY_DENOM = "lamports"
RENT_EXEMPT_AMOUNT = 1000000

LAMPORTS_PER_SOL = 1000000000
_IDL = "idl"
_BYTECODE = "bytecode"
